package handler

import (
	"crypto/hmac"
	"crypto/sha256"
	_ "embed"
	"encoding/base64"
	"fmt"
	"net/http"
	"net/url"
	"time"

	"github.com/labstack/echo/v4"
	"github.com/rs/zerolog/log"
	"gitlab.bad-wolf.fr/optifur/captcha"
	"gitlab.bad-wolf.fr/optifur/captcha-server/config"
	challenge "gitlab.bad-wolf.fr/optifur/captcha-server/services/challenges"
)

// CaptchaHandler handle captcha related requests
type CaptchaHandler interface {
	GetCaptcha(c echo.Context, req GetCaptchaRequest) (*GetCaptchaResponse, error)
	GetCaptchaResource(c echo.Context, req GetCaptchaResourceRequest) error
	ValidateCaptcha(c echo.Context, req ValidateCaptchaRequest) (*ValidateCaptchaResponse, error)
}

// NewCaptchaHandler return a new user handler
func NewCaptchaHandler() CaptchaHandler {
	cfg := config.Load()
	var key []byte
	if cfg.CaptchatKey != nil {
		key = cfg.CaptchatKey.Key
	}
	manager, err := challenge.NewManager()
	if err != nil {
		log.Fatal().Err(err).Msg("unable to init challenge manager")
	}
	captchatCtrl := captcha.NewCaptchaController(manager, key)
	return &captchaHandler{
		captchatCtrl: captchatCtrl,
	}
}

type captchaHandler struct {
	captchatCtrl *captcha.CaptchaController
}

type GetCaptchaRequest struct {
	Type string `query:"type"`
	Lang string `query:"lang"`
}

type GetCaptchaResponse struct {
	Challenge captcha.Challenge `json:"challenge"`
}

func (h captchaHandler) GetCaptcha(c echo.Context, req GetCaptchaRequest) (*GetCaptchaResponse, error) {

	if req.Lang == "" {
		if fb := config.Load().FallbackLang; fb != nil && *fb != "" {
			req.Lang = *fb
		} else {
			return nil, echo.NewHTTPError(http.StatusBadRequest, "field lang missing")
		}
	}

	if req.Type == "" {
		return nil, echo.NewHTTPError(http.StatusBadRequest, "field type missing")
	}

	challenge, err := h.captchatCtrl.NewChallenge(req.Type, req.Lang)
	if err != nil {
		return nil, fmt.Errorf("unable to generate new challenge: %w", err)
	}

	return &GetCaptchaResponse{
		Challenge: *challenge,
	}, nil
}

type GetCaptchaResourceRequest struct {
	ID string `param:"id"`
}

func (h captchaHandler) GetCaptchaResource(c echo.Context, req GetCaptchaResourceRequest) error {

	id, err := url.PathUnescape(req.ID)
	if err != nil {
		return fmt.Errorf("unable to unescape id: %w", err)
	}
	path, err := h.captchatCtrl.ResourcePath(id)
	if err != nil {
		return fmt.Errorf("unable to acces resource: %w", err)
	}
	return c.File(path)
}

type ValidateCaptchaRequest struct {
	Answer []string `json:"answer"`
	Ctrl   string   `json:"ctrl"`
}

type ValidateCaptchaResponse struct {
	Valide          bool    `json:"valide"`
	ValidationToken *string `json:"validationToken,omitempty"`
}

func (h captchaHandler) ValidateCaptcha(c echo.Context, req ValidateCaptchaRequest) (*ValidateCaptchaResponse, error) {
	// First, wait one second to prevent brut-force...
	time.Sleep(time.Second)

	ok, err := h.captchatCtrl.CheckControl(req.Answer, req.Ctrl, 5*time.Minute)
	if err != nil {
		return nil, fmt.Errorf("unable to validate captcha: %w", err)
	}

	res := &ValidateCaptchaResponse{
		Valide: ok,
	}
	cfg := config.Load()
	if ok && cfg.CaptchatSharedKey != nil {
		now := time.Now().Unix()
		h := hmac.New(sha256.New, cfg.CaptchatSharedKey.Key)
		h.Write([]byte(fmt.Sprintf("%d", now)))
		signatureBytes := h.Sum(nil)
		signatureB64 := base64.StdEncoding.EncodeToString(signatureBytes)
		token := fmt.Sprintf("%d.%s", now, signatureB64)

		res.ValidationToken = &token
	}

	return res, nil
}
