package sdk

import (
	"crypto/hmac"
	"crypto/sha256"
	"encoding/base64"
	"fmt"
	"strconv"
	"strings"
	"time"
)

// CheckToken allow checking a token validity.
// Token is composed of a timestamp in second and a signature of said token by
// the captcha server. Key must be the same as the captchatSharedKey in the
// server config
// If lifespan >0, the token is considered valide only if not older than provided lifespan
func CheckToken(token string, key []byte, lifespan time.Duration) (bool, error) {

	s := strings.SplitN(token, ".", 2)
	if len(s) != 2 {
		return false, fmt.Errorf("malformed token")
	}
	ts := s[0]
	signature := s[1]

	h := hmac.New(sha256.New, key)
	h.Write([]byte(ts))
	signatureBytes := h.Sum(nil)
	signatureB64 := base64.StdEncoding.EncodeToString(signatureBytes)
	if signatureB64 != signature {
		return false, nil
	}

	// Signature ok, check duration
	if lifespan > 0 {

		tsInt, err := strconv.Atoi(ts)
		if err != nil {
			return false, fmt.Errorf("unable to parse timestamp: %w", err)
		}
		tokenDate := time.Unix(int64(tsInt), 0)
		if time.Since(tokenDate) > lifespan {
			return false, nil
		}
	}
	return true, nil
}
