package challenge

import (
	"encoding/base64"
	"encoding/json"
	"fmt"
	"math/rand"
	"os"
	"path"
	"strings"

	"gitlab.bad-wolf.fr/optifur/captcha"
	"gitlab.bad-wolf.fr/optifur/captcha-server/config"
	"golang.org/x/exp/slices"
)

func NewManager() (captcha.ChallengeManager, error) {
	cfg := config.Load()

	captchatConfig := ChallengeConfig{}
	if err := json.Unmarshal(cfg.CaptchatConfig.Content, &captchatConfig); err != nil {
		return nil, fmt.Errorf("unable to decode config file: %w", err)
	}
	return &manager{
		captchatConfig: captchatConfig,
	}, nil
}

type manager struct {
	captchatConfig ChallengeConfig
}

func (m *manager) ListChallenges(kind string, lang string) ([]captcha.ChallengeType, error) {
	res := []captcha.ChallengeType{}

	for k, v := range m.captchatConfig.Challenges {

		if strings.EqualFold(v.Kind, kind) && slices.Contains(v.Lang, lang) {
			c := v.ChallengeType
			c.Metadata["name"] = k
			challengeVars := map[string]any{}
			for varName, varInfo := range v.Vars {
				varValue, err := varInfo.Generate()
				if err != nil {
					return nil, fmt.Errorf("unable to generate var: %w", err)
				}
				challengeVars[varName] = varValue
			}

			c.Data = map[string]any{
				"vars": challengeVars,
			}

			res = append(res, c)
		}
	}

	return res, nil
}

func (m *manager) listAssets(resPath string) ([]string, error) {
	res := []string{}
	entries, err := os.ReadDir(resPath)
	if err != nil {
		return nil, fmt.Errorf("unable to list assets: %w", err)
	}

	for _, e := range entries {
		if !e.IsDir() {
			res = append(res, e.Name())
		}
	}
	return res, nil
}

func (m *manager) getAnswerIDs(challenge captcha.ChallengeType, count int, good bool) ([]string, error) {
	challengeName := challenge.Metadata["name"]
	if challengeName == "" {
		return nil, fmt.Errorf("challenge name missing")
	}

	challengeInfo, ok := m.captchatConfig.Challenges[challengeName]
	if !ok {
		return nil, fmt.Errorf("challenge %s not found", challengeName)
	}

	assetPath := path.Join(m.captchatConfig.AssetFolder, challengeInfo.ResourcePath)

	assets, err := m.listAssets(assetPath)
	if err != nil {
		return nil, fmt.Errorf("unable to list assets: %w", err)
	}

	rand.New(CryptoSource{}).Shuffle(len(assets), func(i, j int) { //nolint:gosec
		assets[i], assets[j] = assets[j], assets[i]
	})

	var vars map[string]any
	if data, ok := challenge.Data.(map[string]any); ok {
		vars, ok = data["vars"].(map[string]any)
		if !ok {
			return nil, fmt.Errorf("found invalid vars value in data")
		}
	}

	var matchers []*ResourceMatcher
	if good {
		matchers = challengeInfo.GoodResources
	} else {
		matchers = challengeInfo.BadResources
	}
	res := []string{}
assetsLoop:
	for _, v := range assets {
		for _, m := range matchers {
			ok, err := m.Match(v, vars)
			if err != nil {
				return nil, fmt.Errorf("unable to match asset: %w", err)
			}
			if ok {
				res = append(res, base64.StdEncoding.EncodeToString([]byte(path.Join(assetPath, v))))
				if len(res) == count {
					return res, nil
				}
				continue assetsLoop
			}
		}
	}

	return nil, fmt.Errorf("unable to find requested amount of good answer")
}

func (m *manager) GetGoodAnswerIDs(challenge captcha.ChallengeType, count int) ([]string, error) {
	return m.getAnswerIDs(challenge, count, true)
}

func (m *manager) GetBadAnswerIDs(challenge captcha.ChallengeType, count int) ([]string, error) {
	return m.getAnswerIDs(challenge, count, false)
}

func (m *manager) GetResourcePath(id string) (string, error) {
	idBytes, err := base64.StdEncoding.DecodeString(id)
	if err != nil {
		return "", fmt.Errorf("unable to decode id: %w", err)
	}
	return string(idBytes), nil
}
