package challenge

import (
	"bytes"
	"crypto/rand"
	"encoding/json"
	"errors"
	"fmt"
	"math/big"
	"regexp"
	"text/template"

	"github.com/dop251/goja"
	"github.com/dop251/goja_nodejs/console"
	"github.com/dop251/goja_nodejs/require"
	"github.com/rs/zerolog/log"
	"gitlab.bad-wolf.fr/optifur/captcha"
)

type ChallengeConfig struct {
	AssetFolder string               `json:"assetFolder"`
	Challenges  map[string]Challenge `json:"challenges"`
}

type Challenge struct {
	captcha.ChallengeType
	Lang          []string           `json:"lang"`
	ResourcePath  string             `json:"resourcePath"`
	GoodResources []*ResourceMatcher `json:"goodResources"`
	BadResources  []*ResourceMatcher `json:"badResources"`
	Vars          map[string]Var     `json:"vars"`
}

type ResourceMatcher struct {
	Name               string  `json:"name"`
	ConstraintFunction *string `json:"constraintFunction"`
	Invert             bool    `json:"invert"`
}

func (m *ResourceMatcher) Match(path string, vars map[string]any) (bool, error) {

	tmpl, err := template.New("").Parse(m.Name)
	if err != nil {
		return false, fmt.Errorf("unable to parse template: %w", err)
	}
	dest := bytes.NewBuffer(nil)
	if err := tmpl.Execute(dest, map[string]any{"vars": vars}); err != nil {
		return false, fmt.Errorf("unable to execute template: %w", err)
	}

	re, err := regexp.Compile(dest.String())
	if err != nil {
		return false, fmt.Errorf("unable to compile regex: %w", err)
	}
	doMatch := false
	if m.ConstraintFunction != nil {
		matches := re.FindStringSubmatch(path)
		if len(matches) != 0 {
			vm := goja.New()
			// Enable console for debuging purpose
			new(require.Registry).Enable(vm)
			console.Enable(vm)

			for i, v := range matches {
				if err := vm.Set(fmt.Sprintf("$%d", i), v); err != nil {
					return false, fmt.Errorf("unable to set regex capture value: %w", err)
				}
			}
			if err := vm.Set("$vars", vars); err != nil {
				return false, fmt.Errorf("unable to set regex vars value: %w", err)
			}

			res, err := vm.RunString(*m.ConstraintFunction)
			if err != nil {
				exception := &goja.Exception{}
				if ok := errors.As(err, &exception); !ok {
					return false, fmt.Errorf("unable to run ConstraintFunction: %w", err)
				}
				log.Warn().Err(exception).Msg("js error in ConstraintFunction")
				// JS exception => false
				res = vm.ToValue(false)
			}
			doMatch = res.ToBoolean()
		}
	} else {
		doMatch = re.MatchString(path)
	}
	return doMatch != m.Invert, nil
}

// #### Var handling section ####

// Var is a variable. Its behaviour depends on its type
type Var struct {
	Type      string `json:"type"`
	Generater `json:"-"`
}

// varNoUnmarshal alias to prevent unmarshal infinite loop
type varNoUnmarshal Var

// UnmarshalJSON implements the json.Unmarshaler interface
func (v *Var) UnmarshalJSON(d []byte) error {
	if err := json.Unmarshal(d, (*varNoUnmarshal)(v)); err != nil {
		return fmt.Errorf("unable to unmarshal var type")
	}
	switch v.Type {
	case "enum":
		target := VarEnum{}
		if err := json.Unmarshal(d, &target); err != nil {
			return fmt.Errorf("unable to unmarshal target var: %w", err)
		}
		v.Generater = target
	case "int":
		target := VarInt{}
		if err := json.Unmarshal(d, &target); err != nil {
			return fmt.Errorf("unable to unmarshal target var: %w", err)
		}
		v.Generater = target
	default:
		return fmt.Errorf("unable to unmarshal var: type '%s' unknown", v.Type)
	}
	return nil
}

// MarshalJSON implements the json.Marshaler interface
func (v Var) MarshalJSON() ([]byte, error) {
	type varBis Var
	vRes, err := json.Marshal(varBis(v))
	if err != nil {
		return nil, fmt.Errorf("unable to marshall variable")
	}

	gRes, err := json.Marshal(v.Generater)
	if err != nil {
		return nil, fmt.Errorf("unable to marshall generater")
	}

	res := vRes[:len(vRes)-1]
	res = append(res, ',')
	res = append(res, gRes[1:]...)
	return res, nil
}

// Generater are implementation of var
type Generater interface {
	Generate() (any, error)
}

// VarEnum will chose a value from a list of choice
type VarEnum struct {
	Choices []string `json:"choices"`
}

// Generate implements the Generater interface
func (v VarEnum) Generate() (any, error) {
	maxValue := big.NewInt(int64(len(v.Choices)))
	res, err := rand.Int(rand.Reader, maxValue)
	if err != nil {
		return nil, fmt.Errorf("unable to generate number")
	}
	return v.Choices[res.Int64()], nil
}

// VarInt will chose a random int value in [min, max]
type VarInt struct {
	Min int `json:"min"`
	Max int `json:"max"`
}

// Generate implements the Generater interface
func (v VarInt) Generate() (any, error) {
	maxValue := big.NewInt(int64(v.Max - v.Min + 1))
	res, err := rand.Int(rand.Reader, maxValue)
	if err != nil {
		return nil, fmt.Errorf("unable to generate number")
	}
	return res.Int64() + int64(v.Min), nil
}
