package main_test

import (
	"bytes"
	"context"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"io"
	"net/http"
	"net/http/httptest"
	"os"
	"path"
	"strconv"
	"testing"
	"time"

	_ "embed"

	"github.com/labstack/echo/v4/middleware"
	"github.com/labstack/gommon/log"
	. "github.com/smartystreets/goconvey/convey"
	"gitlab.bad-wolf.fr/optifur/captcha-server/config"
	"gitlab.bad-wolf.fr/optifur/captcha-server/handler"
	"gitlab.bad-wolf.fr/optifur/captcha-server/router"
	"gitlab.bad-wolf.fr/optifur/captcha-server/sdk"
	challenge "gitlab.bad-wolf.fr/optifur/captcha-server/services/challenges"
	configmodel "gitlab.bad-wolf.fr/optifur/config"
)

//go:embed challenges.json
var challengeConfigJSON []byte

func TestCaptchaRequest(t *testing.T) {

	ctx := context.Background()
	Convey("Settup test server", t, func() {

		var challengeConfig challenge.ChallengeConfig
		err := json.Unmarshal(challengeConfigJSON, &challengeConfig)
		So(err, ShouldBeNil)

		tmpDir, err := os.MkdirTemp("", fmt.Sprintf("go_test_%s", time.Now()))
		So(err, ShouldBeNil)

		challengeConfig.AssetFolder = tmpDir
		err = os.MkdirAll(path.Join(tmpDir, "visual/test"), 0700)
		So(err, ShouldBeNil)

		// Here folders are created, create files for each dice value
		for i := 1; i <= 20; i++ {
			for k := 0; k < 4; k++ {
				file, err := os.Create(path.Join(tmpDir, "visual/test", fmt.Sprintf("%d_%d", i, k))) //nolint:gosec
				So(err, ShouldBeNil)

				// Write value as user for testing purpose
				_, err = file.WriteString(fmt.Sprintf("%d", i))
				So(err, ShouldBeNil)

				So(file.Close(), ShouldBeNil)
			}
		}

		challengeConfigJSON, err = json.Marshal(challengeConfig)
		So(err, ShouldBeNil)

		cfg := config.Load()
		cfg.CaptchatConfig.Content = challengeConfigJSON
		cfg.CaptchatSharedKey = new(configmodel.B64KeyFile)
		cfg.CaptchatSharedKey.Key = []byte("potato_key")

		router := router.InitRouter(ctx)
		router.Use(middleware.Logger())
		router.Debug = true
		router.Logger.SetLevel(log.DEBUG)
		router.Logger.SetOutput(os.Stdout)
		router.StdLogger.SetOutput(os.Stdout)

		server := httptest.NewServer(router)
		defer server.Close()

		Convey("Running server", func() {

			Convey("Request captcha", func() {

				// Prepare request with signature
				req, err := http.NewRequest(http.MethodGet, server.URL+"/captcha?type=visual&lang=fr", nil)
				So(err, ShouldEqual, nil)

				// Perform request
				resp, err := http.DefaultClient.Do(req)
				So(err, ShouldEqual, nil)
				defer func() {
					if err := resp.Body.Close(); err != nil {
						t.Error(err)
					}
				}()
				So(resp.StatusCode, ShouldEqual, http.StatusOK)

				c := handler.GetCaptchaResponse{}
				So(json.NewDecoder(resp.Body).Decode(&c), ShouldBeNil)
				So(c.Challenge.ChallengeType.Kind, ShouldEqual, "visual")

				diceValue, ok := c.Challenge.ChallengeType.Data.(map[string]any)["vars"].(map[string]any)["diceValue"].(float64)
				So(ok, ShouldBeTrue)
				So(diceValue, ShouldBeBetweenOrEqual, 3, 18)

				Convey("Validate captcha", func() {

					// Search for good values
					goodIDs := []string{}
					for _, id := range c.Challenge.ResourcesID {
						req, err := http.NewRequest(http.MethodGet, server.URL+"/captcha/resources/"+id, nil)
						So(err, ShouldEqual, nil)

						// Perform request
						resp, err := http.DefaultClient.Do(req)
						So(err, ShouldEqual, nil)
						defer func() {
							if err := resp.Body.Close(); err != nil {
								t.Error(err)
							}
						}()
						So(resp.StatusCode, ShouldEqual, http.StatusOK)

						bodyBytes, err := io.ReadAll(resp.Body)
						So(err, ShouldEqual, nil)

						resValue, err := strconv.Atoi(string(bodyBytes))
						So(err, ShouldEqual, nil)

						if resValue >= int(diceValue) {
							goodIDs = append(goodIDs, id)
						}
					}

					// Prepare request with signature

					validateReq := handler.ValidateCaptchaRequest{
						Ctrl:   c.Challenge.Controle,
						Answer: goodIDs,
					}

					reqBody, err := json.Marshal(validateReq)
					So(err, ShouldEqual, nil)

					req, err := http.NewRequest(http.MethodPost, server.URL+"/captcha", bytes.NewBuffer(reqBody))
					So(err, ShouldEqual, nil)

					req.Header.Add("content-type", "application/json")

					// Perform request
					resp, err := http.DefaultClient.Do(req)
					So(err, ShouldEqual, nil)
					defer func() {
						if err := resp.Body.Close(); err != nil {
							t.Error(err)
						}
					}()
					So(resp.StatusCode, ShouldEqual, http.StatusOK)

					c := handler.ValidateCaptchaResponse{}
					So(json.NewDecoder(resp.Body).Decode(&c), ShouldBeNil)
					So(c.Valide, ShouldBeTrue)
					So(c.ValidationToken, ShouldNotBeNil)

					valid, err := sdk.CheckToken(*c.ValidationToken, []byte("potato_key"), 0)
					So(valid, ShouldBeTrue)
					So(err, ShouldEqual, nil)

					valid, err = sdk.CheckToken(fmt.Sprintf("%d.%s", time.Now().Unix(), base64.StdEncoding.EncodeToString([]byte("fried potatoes"))), []byte("potato_key"), 0)
					So(valid, ShouldBeFalse)
					So(err, ShouldEqual, nil)
				})

				Convey("Validate captcha wrong answer", func() {

					// Search for bad values or mix
					badIDs := []string{}
					for _, id := range c.Challenge.ResourcesID {
						req, err := http.NewRequest(http.MethodGet, server.URL+"/captcha/resources/"+id, nil)
						So(err, ShouldEqual, nil)

						// Perform request
						resp, err := http.DefaultClient.Do(req)
						So(err, ShouldEqual, nil)
						defer func() {
							if err := resp.Body.Close(); err != nil {
								t.Error(err)
							}
						}()
						So(resp.StatusCode, ShouldEqual, http.StatusOK)

						bodyBytes, err := io.ReadAll(resp.Body)
						So(err, ShouldEqual, nil)

						resValue, err := strconv.Atoi(string(bodyBytes))
						So(err, ShouldEqual, nil)

						if resValue <= int(diceValue) {
							badIDs = append(badIDs, id)
						}
					}

					// Prepare request with signature

					validateReq := handler.ValidateCaptchaRequest{
						Ctrl:   c.Challenge.Controle,
						Answer: badIDs,
					}

					reqBody, err := json.Marshal(validateReq)
					So(err, ShouldEqual, nil)

					req, err := http.NewRequest(http.MethodPost, server.URL+"/captcha", bytes.NewBuffer(reqBody))
					So(err, ShouldEqual, nil)

					req.Header.Add("content-type", "application/json")

					// Perform request
					resp, err := http.DefaultClient.Do(req)
					So(err, ShouldEqual, nil)
					defer func() {
						if err := resp.Body.Close(); err != nil {
							t.Error(err)
						}
					}()
					So(resp.StatusCode, ShouldEqual, http.StatusOK)

					c := handler.ValidateCaptchaResponse{}
					So(json.NewDecoder(resp.Body).Decode(&c), ShouldBeNil)
					So(c.Valide, ShouldBeFalse)
					So(c.ValidationToken, ShouldBeNil)
				})

			})
		})
	})
}
